
from pca9685_driver.device import Registers

class FakeSMBus(object):
    def __init__(self, bus_number):
        self.values = {
            Registers.MODE_1: 0b00010001,
            8: 255,
            9: 7,
        }
        self.wrote_values = []
        self.bus_number = bus_number

    def write_byte_data(self, address, register, value):
        self.values[register] = value
        self.wrote_values.append((register, value))

    def read_byte_data(self, address, register):
        return self.values[register]



import smbus
import logging
import glob
import re

logger = logging.getLogger(__name__)

class DeviceException(Exception):
    pass

class Registers(object):
    MODE_1 = 0x00
    MODE_2 = 0x01
    LED_STRIP_START = 0x06 # LED0 ON Low Byte
    PRE_SCALE = 0xFE

class Mode1(object):
    RESTART = 7
    EXTCLK = 6
    AI = 5
    SLEEP = 4
    SUB1 = 3
    SUB2 = 2
    SUB3 = 1
    ALLCALL = 0

class Mode2(object):
    INVRT = 4
    OCH = 3
    OUTDRV = 2
    OUTNE_1 =1
    OUTNE_0 = 0

def value_low(val):
    return val & 0xFF

def value_high(val):
    return (val >> 8) & 0xFF

class Device(object):

    ranges = dict(
        pwm_frequency = (24, 1526),
        led_number = (0, 15),
        led_value = (0, 4095),
        register_value = (0, 255),
    )

    def __init__(self, address, bus_number = None, bus_interface_factory = smbus.SMBus, glober = glob.glob):
        """Creates an interface to PCA9685 device

        :param address: the I2C address of the device. Check the addressed with `i2cdetect -y 1`
        :param bus_number: the number of the I2C bus in the linux machine. See /dev/i2c-*
        :param bus_interface_factory: bus class factory, used for unit tests
        :param glober: for search in file system, used for unit tests
        """

        if bus_number is None:
            bus_list = Device.get_i2c_bus_numbers(glober)
            if len(bus_list) < 1:
                raise DeviceException("Cannot determine I2C bus number")
            bus_number = bus_list[0]

        self.__address = address
        self.__bus = bus_interface_factory(bus_number)
        self.__oscillator_clock = 25000000

    @staticmethod
    def get_i2c_bus_numbers(glober = glob.glob):
        """Search all the available I2C devices in the system"""
        res = []
        for device in glober("/dev/i2c-*"):
            r = re.match("/dev/i2c-([\d]){1,2}", device)
            res.append(int(r.group(1)))
        return res

    @property
    def mode_1(self):
        """Returns the Mode 1 register value"""
        return self.read(Registers.MODE_1)

    @property
    def bus(self):
        """Returns the bus instance"""
        return self.__bus

    def get_led_register_from_name(self, name):
        """Parse the name for led number

        :param name: attribute name, like: led_1
        """
        res = re.match('^led_([0-9]{1,2})$', name)
        if res is None:
            raise AttributeError("Unknown attribute: '%s'" % name)
        led_num = int(res.group(1))
        if led_num < 0 or led_num > 15:
            raise AttributeError("Unknown attribute: '%s'" % name)
        return self.calc_led_register(led_num)

    def calc_led_register(self, led_num):
        """Calculate register number for LED pin

        :param led_num: the led number, typically 0-15
        """
        start = Registers.LED_STRIP_START + 2
        return start + (led_num * 4)

    def __check_range(self, type, value):
        range = self.ranges[type]
        if value < range[0]:
            raise DeviceException("%s must be greater than %s, got %s" % (type, range[0], value))
        if value > range[1]:
            raise DeviceException("%s must be less than %s, got %s" % (type, range[1], value))

    def set_pwm(self, led_num, value):
        """Set PWM value for the specified LED

        :param led_num: LED number (0-15)
        :param value: the 12 bit value (0-4095)
        """
        self.__check_range('led_number', led_num)
        self.__check_range('led_value', value)

        register_low = self.calc_led_register(led_num)

        self.write(register_low, value_low(value))
        self.write(register_low + 1, value_high(value))

    def __get_led_value(self, register_low):
        low = self.read(register_low)
        high = self.read(register_low + 1)
        return low + (high * 256)

    def get_pwm(self, led_num):
        """Generic getter for all LED PWM value"""
        self.__check_range('led_number', led_num)
        register_low = self.calc_led_register(led_num)
        return self.__get_led_value(register_low)

    def __getattr__(self, name):
        """Generic getter property handler for all LED PWM value"""
        register_low = self.get_led_register_from_name(name)
        return self.__get_led_value(register_low)

    def sleep(self):
        """Send the controller to sleep"""
        logger.debug("Sleep the controller")
        self.write(Registers.MODE_1, self.mode_1 | (1 << Mode1.SLEEP))

    def wake(self):
        """Wake up the controller"""
        logger.debug("Wake up the controller")
        self.write(Registers.MODE_1, self.mode_1 & (255 - (1 << Mode1.SLEEP)))

    def write(self, reg, value):
        """Write raw byte value to the specified register

        :param reg: the register number (0-69, 250-255)
        :param value: byte value
        """
        # TODO: check reg: 0-69, 250-255
        self.__check_range('register_value', value)
        logger.debug("Write '%s' to register '%s'" %  (value, reg))
        self.__bus.write_byte_data(self.__address, reg, value)

    def read(self, reg):
        """Read data from register

        :param reg: the register number (0-69, 250-255)
        """
        return self.__bus.read_byte_data(self.__address, reg)

    def calc_pre_scale(self, frequency):
        """Calculate the controller's PRE_SCALE value, specified by the PCA9685 datasheet

        :param frequency: source frequency value
        """
        return int(round(self.__oscillator_clock / (4096.0 * frequency)) - 1)

    def set_pwm_frequency(self, value):
        """Set the frequency for all PWM output

        :param value: the frequency in Hz
        """
        self.__check_range('pwm_frequency', value)
        reg_val = self.calc_pre_scale(value)
        logger.debug("Calculated prescale value is %s" % reg_val)
        self.sleep()
        self.write(Registers.PRE_SCALE, reg_val)
        self.wake()

    def calc_frequency(self, prescale):
        """Calculate the frequency by the controller's prescale, specified by the PCA9685 datasheet

        :param prescale: the prescale value of the controller
        """
        return int(round(self.__oscillator_clock / ((prescale + 1) * 4096.0)))

    def get_pwm_frequency(self):
        """Gets the frequency for all PWM output"""

        return self.calc_frequency(self.read(Registers.PRE_SCALE))



import unittest

from pca9685_driver import Device, DeviceException

from .fake_smbus import FakeSMBus

class TestDevice(unittest.TestCase):

    def test_search_for_i2c_devices(self):
        # Arrange
        def glob(pattern):
            return ['/dev/i2c-8']

        # Act
        dev = Device(0, None, FakeSMBus, glob)

        # Assert
        self.assertEqual(dev.bus.bus_number, 8)

    def test_calc_pre_scale(self):
        # Arrange
        dev = Device(0, 0, FakeSMBus)

        # Act & Assert
        self.assertEqual(dev.calc_pre_scale(200), 30)
        self.assertEqual(dev.calc_pre_scale(1000), 5)

    def test_set_pwm_frequency(self):
        # Arrange
        dev = Device(0, 0, FakeSMBus)

        # Act
        dev.set_pwm_frequency(200)

        # Assert
        self.assertEqual(dev.bus.wrote_values[0], (0, 17))
        self.assertEqual(dev.bus.wrote_values[1], (254, 30))
        self.assertEqual(dev.bus.wrote_values[2], (0, 1))

    def test_set_pwm_frequency_wrong_input(self):
        # Arrange
        dev = Device(0, 0, FakeSMBus)

        # Act & Assert
        with self.assertRaises(DeviceException):
            dev.set_pwm_frequency(2000)
        with self.assertRaises(DeviceException):
            dev.set_pwm_frequency(10)

    def test_led_number_getter(self):
        # Arrange
        dev = Device(0, 0, FakeSMBus)

        # Act & Assert
        self.assertEqual(dev.led_0, 2047) # default value in FakeSMBus

    def test_led_number_wrong_getter(self):
        # Arrange
        dev = Device(0, 0, FakeSMBus)

        # Act & Assert
        with self.assertRaises(AttributeError):
            self.assertEqual(dev.led_42, 2047)
        with self.assertRaises(AttributeError):
            self.assertEqual(dev.teve, 2047)

    def test_set_pwm_value(self):
        # Arrange
        dev = Device(0, 0, FakeSMBus)

        # Act
        dev.set_pwm(4, 1042)

        # Assert
        self.assertEqual(dev.bus.values[24], 18)
        self.assertEqual(dev.bus.values[25], 4)

    def test_calc_frequency(self):
        # Arrange
        dev = Device(0, 0, FakeSMBus)

        # Act & Assert
        self.assertEqual(dev.calc_frequency(30), 197)
        self.assertEqual(dev.calc_frequency(5), 1017)

    def test_get_pwm_frequency(self):
        # Arrange
        dev = Device(0, 0, FakeSMBus)

        # Act
        dev.set_pwm_frequency(197)

        # Assert
        self.assertEqual(dev.get_pwm_frequency(), 197)



from setuptools import setup, find_packages

setup(
    name = "PCA9685-driver",
    description = "Driver module for PCA9685 pwm (LED/Servo) controller",
    version = "1.1.0",
    author = 'Lajos Santa',
    author_email = 'santa.lajos@coldline.hu',
    url = 'https://github.com/voidpp/PCA9685-driver.git',
    install_requires = [
        "smbus-cffi==0.5.1",
    ],
    include_package_data = True,
    packages = find_packages(),
)
